library('pracma')
library('mvtnorm')
library('Matrix')
library(extraDistr)
library(coda)

Gibbs_sampler = function(K, L, R, group_sizes, knots, deg, niter, neff, seed_info, X = NULL, burnin = 0, a_eta = 0, b_eta = 0,
                         a_tau=0, b_tau=0, a_rho = 0,b_rho = 0, prior_mean_beta = NULL, prior_precision_beta = NULL,
                         m_0 = 0, v_0 = 0, init_mean = NULL, init_Cov= NULL, print_progress = FALSE, sample_effective = FALSE,
                         true_params = list(), previous_results = NULL, test_convg = T, save_intermediate = F, write_=F, name = '', M = Inf
  ){
  # Gibbs sampler for posterior sampling under type-2 interval-censored survival data, assuming a 2-level nested multivariate probit model with 
  # monotone spline baseline function h. Level 1 is balanced nested (e.g. event times per person) and level 2 is unbalanced nested (e.g. treatment groups).
  # K is the number of events grouped in level 1, L and R are the vectors of left and right interval
  # endpoints respectively, group_sizes is a vector of group sizes for level 2, knots are the spline knots for the baseline, deg is the spline degree of the baseline
  # niter is the (effective/absolute) number of Gibbs sampler iterations. X is the covariate matrix, burnin is the burnin period (# iterations). a_eta and b_eta 
  # are the prior shape and scale parameters for the inverse gamma distribution for the hierarchical eta parameter for the spline coefficients, respectively.
  # a_tau and b_tau are the prior shape and scale parameters for the inverse gamma distribution for the covariance parameter tau for level 2, respectively.
  # A similar definition holds for a_rho b_rho, where rho is the covariance parameter for the first level. prior_mean_beta and prior_precision_beta are the prior mean 
  # vector and precision matrix for the multivariate normally distributed beta parameters. m_0,v_0 are the prior mean and precision for the normally distriubted gamma_0 (intercept for spline function).
  # init_mean and init_Cov denote the initial mean and covariance matrix for the multivariate normal distribution with which the initial values for beta, gamma are sampled ( gamma[-1] is set to absolute value).
  # print_progress is a setting to print the intermediate samples, sample_effective=T makes the algorithm run until an effective sample size of niter is reached.
  # true_params is a list of the true parameter values (if they are known) used for printing. prior_iterations is a possible matrix of previously sampled values of beta,tau,rho,gamma (in that order)
  # test_convg specifies whether the Geweke diagnostic should be passed before initializing the sample. save_intermediate denotes whether the method stores the intermediate values of theta, the matrix of samples.
  
  # set seed
  assertthat::assert_that(!is.null(seed_info$seed))
  set.seed(seed_info$seed)
  seed_info$randomseed = .Random.seed 

  # write empty progress file
  write('',paste0('out_', name, ".txt"))

  # setting constants for sampler
  n   = length(group_sizes)   # num groups
  nullval_X = is.null(X)
  if(nullval_X){
    X = zeros(K*sum(group_sizes),2)
    p=0
  }else{
     p   = dim(X)[2]             # num covariates
     X   = -X                    # flipping the time axis according to Lin Wang article 
  }
 
  # obtaining matrix sigma (such that sigma%*%gamma = alpha(t_i) from Lin Wang article), 
  T_LW               = L
  T_LW[T_LW == 0]    = R[T_LW == 0]
  sigma              = get_spline_matrix_points(T_LW, knots = knots, deg = deg)
  sigma[is.infinite(T_LW),] = 0
  if(!is.null(X)){
    X[is.infinite(T_LW),]= 0
  }

  #calculating matrices with which to transform the data later on
  cugs                  = c(0, cumsum(group_sizes))
  cluster_indc_mat      = bdiag(lapply(group_sizes, function(gs){ones(gs*K,1)}))
  subject_indc_mat      = bdiag(lapply(seq(sum(group_sizes)), function(i){ones(K,1)}))
  sum_sigma_clusters    = t(cluster_indc_mat)%*%sigma   
  sum_sigma_subjects    = t(subject_indc_mat)%*%sigma
  sum_X_clusters        = t(cluster_indc_mat)%*%X   
  sum_X_subjects        = t(subject_indc_mat)%*%X
  
  #determining matrices with which one can sample tau,rho
  Helmerttransformmat   = bdiag(lapply(group_sizes, function(gs){kron(HelmertMatrix(gs), HelmertMatrix(K))})) 
  start_cluster_indices = cugs[seq(n)]*K+1                                                #first indices in cluster, elements containing info about tau if Helmert transformed
  other_subject_indices = setdiff(seq(1,sum(group_sizes)*K,K), start_cluster_indices)     #subject indices, excluding the above indices, elements contain info about rho if Helmert transformed
  suffstat_tau_mat      = Helmerttransformmat[start_cluster_indices,]
  suffstat_rho_mat      = Helmerttransformmat[other_subject_indices,]
  
  # obtaining matrices delta_left and delta_right such that delta_left%*%gamma = -alpha(R) + alpha(t_i).
  L_splines          = get_spline_matrix_points(L,knots, deg)
  R_splines          = get_spline_matrix_points(R, knots, deg)
  delta_left         = -R_splines + sigma
  delta_right        = -L_splines + sigma
  num_funcs          = dim(delta_left)[2] - 1
  
  #obtaining delta left where there is no censoring, and furthermore indicators when this is nonzero (used in gamma sampling)
  IndcIntvCens       = L>0&is.finite(R)           
  delta_left_noncens = delta_left[IndcIntvCens,]             
  IndcDeltaNonzero   = delta_left_noncens < 0
  
  if(!nullval_X){
    #obtaining prior parameters for beta if X != NULL
    prior_mean_beta       = if(is.null(prior_mean_beta)){zeros(p,1)}else{prior_mean_beta}
    prior_precision_beta  = if(is.null(prior_precision_beta)){zeros(p,p)}else{prior_precision_beta}
  }
  
  if(!is.null(previous_results)){ # continue from last state if prior iterations are given

    tau_hat   = previous_results$current_state$tau_hat
    rho_hat   = previous_results$current_state$rho_hat
    beta_hat  = previous_results$current_state$beta_hat
    gamma_hat = previous_results$current_state$gamma_hat
    eta_hat   = previous_results$current_state$eta_hat
    Z_hat     = previous_results$current_state$Z_hat
    
    #resume sampling with old parameters
    theta          = previous_results$MCMC_results
    absolute_iter  = dim(theta)[[1]]
    burnin         = 0
    num_MC_samples = 0
    convgflag      = T
    test_convg     = F
  }else{
    #sampling initial parameters as a multivariate normal vector
    num_params      = p+num_funcs+1
    init_mean       = if(is.null(init_mean)){c(zeros(p+1, 1), 0.1*ones(num_funcs,1))}else{init_mean}
    init_Cov        = if(is.null(init_Cov)){zeros(num_params, num_params)}else{init_Cov}
    init_params     = rmvnorm(1, mean = init_mean, sigma = init_Cov)
    
    if(!nullval_X){
      beta_hat        = init_params[seq(p)]
      gamma_hat       = append(init_params[p+1], abs(init_params[seq(p+2, p+num_funcs+1)]))
    }else{
      beta_hat        = c(0,0)
      gamma_hat       = append(init_params[1], abs(init_params[seq(2, num_funcs+1)]))
    }
    tau_hat         = 0
    rho_hat         = 0
    
    mu              = X%*%beta_hat  + sigma%*%gamma_hat
    Z_hat           = rtnorm(length(mu), a = delta_left%*%gamma_hat, b = delta_right%*%gamma_hat, mean = mu)
    eta_hat         = 1
  
    #init
    theta    = c()
    absolute_iter  = 0
    num_MC_samples = 0
    convgflag = F}
  
  while((num_MC_samples < neff*sample_effective)|absolute_iter-burnin<=niter){
    if(print_progress){
      tic()
    }
    
    #sampling Z
    Z_hat =sample_Z(group_sizes, K, Z_hat, X%*%beta_hat + sigma%*%gamma_hat, delta_left%*%gamma_hat, delta_right%*%gamma_hat, tau_hat, rho_hat,n)
    
    # sampling beta 
    Z_hat_shifted      = Z_hat - sigma%*%gamma_hat
    subject_coef       = -rho_hat/(K*rho_hat + 1)
    clustercoef_matrix = Diagmat(-tau_hat/((K*rho_hat + 1)*(1+K*(rho_hat + group_sizes*tau_hat))))
    if(!nullval_X){
    beta_hat           = sample_beta(X, Z_hat_shifted, clustercoef_matrix, sum_X_clusters,
                                     cluster_indc_mat,subject_coef, sum_X_subjects, subject_indc_mat, prior_mean_beta, prior_precision_beta, M )
    }
    
    # sampling gamma
    gamma_hat = sample_gamma(gamma_hat, Z_hat, sigma, X, beta_hat,  num_funcs, IndcDeltaNonzero, clustercoef_matrix, sum_sigma_clusters, cluster_indc_mat,
                             subject_coef,sum_sigma_subjects,subject_indc_mat,IndcIntvCens, delta_left_noncens, eta_hat, m_0, v_0, M=M)
    
    # sampling eta
    #eta_hat = rgamma(1, a_eta + num_funcs, b_eta + sum(gamma_hat[seq(2, num_funcs+1)]))
    p_ub = pgamma(M, a_eta + num_funcs, b_eta + sum(gamma_hat[seq(2, num_funcs+1)]))
    eta_hat = qgamma(runif(1)*p_ub, a_eta + num_funcs, b_eta + sum(gamma_hat[seq(2, num_funcs+1)]))
    
    # sampling tau and rho
    V_hat         = Z_hat - X%*%beta_hat - sigma%*%gamma_hat
    suffstats_tau = as.numeric(suffstat_tau_mat%*%V_hat)
    suffstats_rho = as.numeric(suffstat_rho_mat%*%V_hat)
    res = sample_covariances( suffstats_tau,suffstats_rho, tau_hat, rho_hat, group_sizes, K, a_tau, b_tau, a_rho, b_rho)
    tau_hat = res[[1]]
    rho_hat = res[[2]]

    # appending results
    if(nullval_X){
      totparams = c( tau_hat, rho_hat, gamma_hat)
    }else{
      totparams = c(beta_hat, tau_hat, rho_hat, gamma_hat)
    }
    if(print_progress){ #print
        print(paste('(effective) sample size:', num_MC_samples))
      print(paste("num iter",absolute_iter))
      print(toc())
      if(length(theta)>0){
        if(dim(theta)[2]>100){
        print(abs(geweke.diag(theta[, seq(p+2)], frac1 = 0.2, frac2 = 0.2)[[1]]))
        }}
        print(paste('convergence:', convgflag))
        if(nullval_X){
          print(paste('Sampled paramts:', toString(round(c( tau_hat, rho_hat), 2))))
        }else{
                 print(paste('Sampled paramts:', toString(round(c(beta_hat, tau_hat, rho_hat), 2))))
        }
 
      if(!isempty(true_params)){
        print(paste('True parameters:', toString(round(c(true_params$beta_true, true_params$tau_true, true_params$rho_true),2))))
      }
    }
    
    absolute_iter = absolute_iter + 1
    
    if(absolute_iter > burnin){
      theta = rbind(theta, totparams)
      
      if(test_convg & !convgflag & mod(dim(theta)[1],500)==0 ){
        convgflag = all(abs(geweke.diag(theta[, seq(p+2)], frac1 = 0.2, frac2 = 0.2)[[1]])<1.96)
        if(! convgflag){
          theta = c()
          absolute_iter = burnin + 1
        }
      }
      if(write_){
         Lines = (readLines(paste0('out_', name, ".txt")))
      
         STR = paste0("num_MC_samples: ", num_MC_samples)
         STR = c(STR,paste0("sample_effective: ", sample_effective))
      
         if(test_convg){
           STR = c(STR, paste0("convgflag = ",convgflag))
         }
         STR = c(STR, paste(sapply(as.numeric(totparams), function(x){toString(round(x,2))})))
         if(!isempty(true_params)){
           STR = c(STR, paste(sapply(as.numeric(true_params), function(x){toString(round(x,2))})))
         }
         Lines = c(Lines, STR)
         writeLines(Lines, paste0('out_', name, ".txt") )
    }
      
      if(sample_effective & (convgflag | !test_convg)){
        if(mod(dim(theta)[1],250)==0){
        if(save_intermediate){ #store the matrix of samples
            theta2= theta
            if(nullval_X){
              colnames(theta2) = c( 'tau_hat','rho_hat', sapply(seq(num_funcs+1), function(i){paste('gamma_',toString(i),"_hat", sep = '')}))
            }else{
              colnames(theta2) = c(sapply(seq(p), function(i){paste('beta_', toString(i),"_hat",sep = '')}), 'tau_hat','rho_hat', sapply(seq(num_funcs+1), function(i){paste('gamma_',toString(i),"_hat", sep = '')}))
            }
   	    simresults = list(MCMC_results = theta2, current_state = mget(c('Z_hat', 'beta_hat', 'tau_hat', 'rho_hat', 'eta_hat', 'gamma_hat')),
                              input_parameters = mget(c('a_eta', 'b_eta','a_tau', 'b_tau', 'a_rho', 'b_rho', 'prior_mean_beta', 'prior_precision_beta', 'm_0', 'v_0', 'init_mean', 'init_Cov')),
                              data = mget(c('K', 'L', 'R', 'X','group_sizes', 'knots', 'deg')), seed_info = seed_info, convgflag = convgflag)
            
            save(simresults, file = paste0('intermediate_MCMC_output_',name,'.Rds'))}
  }
          param_interest =theta[,seq(p+2)]
          num_MC_samples = min(effectiveSize(param_interest))} #calculating effective sample size from autocorrelation
      }else{
        if(convgflag | !test_convg){
        num_MC_samples = num_MC_samples+1}
      }
    }
  
  #adding column names to theta
  if(nullval_X){
    colnames(theta) = c( 'tau_hat','rho_hat', sapply(seq(num_funcs+1), function(i){paste('gamma_',toString(i),"_hat", sep = '')}))
  }else{
    colnames(theta) = c(sapply(seq(p), function(i){paste('beta_', toString(i),"_hat",sep = '')}), 'tau_hat','rho_hat', sapply(seq(num_funcs+1), function(i){paste('gamma_',toString(i),"_hat", sep = '')}))
  }
  simresults = list(MCMC_results = theta, current_state = mget(c('Z_hat', 'beta_hat', 'tau_hat', 'rho_hat',  'eta_hat', 'gamma_hat')),
                    input_parameters = mget(c('a_eta', 'b_eta','a_tau', 'b_tau', 'a_rho', 'b_rho', 'prior_mean_beta', 'prior_precision_beta', 'm_0', 'v_0', 'init_mean', 'init_Cov')),
                    data = mget(c('K', 'L', 'R', 'X','group_sizes', 'knots', 'deg')), seed_info = seed_info, convgflag = convgflag)
  return(simresults )
}

