library('pracma')
library(extraDistr)
sample_Z = function(group_sizes, K, Z_vec, mu, lb, ub, tau, rho,n){
  #subtract mean
  Z_vec = Z_vec - mu
  lb    = lb    - mu
  ub    = ub    - mu
  maxGS = max(group_sizes)
  
  if(n>1){
    #putting event vectors and bounds in a matrix where column ~ patient, in order to make updating efficient
    i_clusters   = append(0, cumsum(group_sizes))
    cluster_mat  = NA*ones(maxGS*K,n)
    lb_mat       = zeros(maxGS*K,n)
    ub_mat       = 1e-10*ones(maxGS*K,n)
          
    for(j in seq(n)){
      i_cluster                            = seq(i_clusters[j]*K+1, i_clusters[j+1]*K) #the full cluster indices
      cluster_mat[seq(group_sizes[j]*K),j] = Z_vec[i_cluster]
      ub_mat[seq(group_sizes[j]*K),j]      = ub[i_cluster]
      lb_mat[seq(group_sizes[j]*K),j]      = lb[i_cluster]
    }
    
    indc_removal = is.na(matrix(cluster_mat, ncol = 1))
    cluster_mat[is.na(cluster_mat)] = 0
    
    # determining the sums of Z in each cluster/for each subject
    cluster_sums  = apply(cluster_mat, 2, sum)
    subject_sums  = zeros(maxGS,n)
    
    if(K>1){
      for(i in seq(maxGS)){
        subject_sums[i,] = apply(cluster_mat[seq((i-1)*K+1, i*K), ], 2, sum)
      }}else{
        subject_sums = cluster_mat
      } 
  }else{
    lb_mat = lb
    ub_mat = ub
    cluster_mat  = Z_vec
    cluster_sums = sum(Z_vec)
    subject_sums = kron(diag(group_sizes), ones(1,K))%*%Z_vec 
    indc_removal = ones(length(Z_vec),1)*FALSE
  }
  
  #setting constants depending on rho,tau, K and the group sizes
  h             = rho + tau*(1+K*rho)/(1+K*(rho + (group_sizes-1)*tau))
  subject_coefs = h/((K-1)*h+1)
  cond_sd       = sqrt((K*h+1)/((K-1)*h+1))
  cluster_coefs = tau/(1+K*(rho + (group_sizes - 1)*tau))
  
  #updating Z iteratively
  for(i in seq(maxGS)){
    i_subject         = seq((i-1)*K+1, i*K)                 #indices of i-th subject entries for each cluster
    subject_sums_cons = get_row(subject_sums, i)            # the sums of the Z-values for the i-th subject in each cluster
    cluster_sums      = cluster_sums - subject_sums_cons    # cluster sums without the sums for this subject (used for updates)
    
    
    for(j in seq(K)){
      #for each event j for subject i, get the corresponding Z values, subtract from subject sum, determine the conditional mean and sample truncated normal
      Z_event             = get_row(cluster_mat, i_subject[j])
      subject_sums_cons   = subject_sums_cons - Z_event
      cond_mean           = cluster_coefs*cluster_sums + subject_coefs*(subject_sums_cons - (K-1)*cluster_coefs*cluster_sums)
      Z_event             = rtnorm_custom(length(Z_event), a = get_row(lb_mat, i_subject[j]), b = get_row(ub_mat, i_subject[j]), mean = cond_mean, sd = cond_sd)
      Z_event             = Z_event*as.numeric(group_sizes>=i) #fix dummy elements to zero
      
      #bookkeeping
      subject_sums_cons = subject_sums_cons + Z_event
      cluster_mat       = set_row(cluster_mat, i_subject[j], Z_event)
    }
    cluster_sums     = cluster_sums + subject_sums_cons
    subject_sums     = set_row(subject_sums, i, subject_sums_cons)
  }
  
  #transforming the updated matrix of Z values to a vector
  Z_vec = matrix(cluster_mat, ncol = 1)
  Z_vec = Z_vec[!indc_removal]        # remove zeros
  Z_vec = Z_vec + mu             # adding mean
  return(Z_vec)
}




if (sys.nframe() == 0){ # if name == main
  source('utils.R')
  n           = 2             # number of clusters
  Egs         = 2             # expected cluster size
  K           = 2             # number of events per subject

  cluster_sizes  = c(2,3)
  sumgs          = sum(cluster_sizes)
  rho            = runif(1, -1/K , 1)
  tau            = runif(1, -1/(K*max(cluster_sizes)) - rho/max(cluster_sizes), 1)
  mu             = zeros(K*sumgs, 1)
  lb             = runif(K*sumgs, -5, -3)
  ub             = runif(K*sumgs, 3, 5)
  Z_vec          = rnorm(K*sumgs)

  #obtain a MCMC sample from Z distributed multivariate normally in clusters with mean zero and cov I + tau + rho*kron(diag(cluster_size), ones(k,k))
  M = 10000
  burnin = 5000
  Z_sample = c()
  for( i in seq(M + burnin)){
    Z_vec =  sample_Z(cluster_sizes, K, Z_vec, mu, lb, ub, tau, rho,n)
    Z_sample = rbind(Z_sample, t(Z_vec))
  }
  Z_sample = Z_sample[seq(burnin, M),]

  #print the true and estimated cov matrices
  C     = cov(Z_sample)
  trueC = bdiag(sapply(cluster_sizes, function(cs){diag(cs*K)+tau+rho*kron(diag(cs), ones(K,K))}))
  print((round(C,1)))
  print(round(trueC, 1))

  #check that no boundaries are crossed
  print(!any((Z_vec>ub)|(any(Z_vec<lb))))
}
