library('pracma')
library('mvtnorm')
library(Matrix)

sample_beta = function(X, Y, clustercoef_matrix, sum_X_clusters, cluster_indc_mat,
                          subject_coef, sum_X_subjects, subject_indc_mat, mu_0, Lambda_0,M=Inf){
  # sample beta parameters a posteriori in a 2-way nested multivariate regression model where Y_i = X_i%*%beta + E_i for each cluster i.
  # It is assumed that E_i ~ N(0, I_{n_1*n_{2i}} + tau_1(I_{n_{2i}} \otimes J_{n_1}) + tau_2 J_{n_1*n_{2i}})
  # X,Y are the (row-wise) concatenations of X_i and Y_i. clustercoef_matrix is a diagonal matrix with \rho_{2i} on the diagonal (from the article)
  # sum_X_clusters is a matrix with rows equal to the sum of covariates for all observations in each layer of level 2.
  # cluster_indc_mat is a matrix whith n_1*n_{2i} rows and n_0 (number of layers of level 2) columns containing either 0 or 1 denoting group membership for the observation belonging to that row.
  # subject_coef, sum_X_subjects and subject_indc_mat have similar definitions for layer 1 where there now only is one rho_1 due to balancedness. 
  # mu_0 and Lambda_0 are the prior mean and precision matrix for the multivarate normal prior for beta. M is a truncation of the support of the prior/posterior (to use uniform proper prior), e.g. M=100
  # would mean that we truncate the support to [-100,100]^p.
  
  #determine posterior covariance matrix Sigma2 and mean mu2
  Sigma2     = as.matrix(solve(t(X)%*%(X) + t(sum_X_clusters)%*%clustercoef_matrix%*%sum_X_clusters + 
                                 subject_coef*t(sum_X_subjects)%*%sum_X_subjects + Lambda_0))

  mu2        = Sigma2%*%(Lambda_0%*%mu_0 + (t(X) + t(sum_X_clusters)%*%clustercoef_matrix%*%t(cluster_indc_mat) +
                                                     subject_coef*t(sum_X_subjects)%*%t(subject_indc_mat))%*%Y)
  
  #sample beta and return
  beta_hat   = t(rtmvnorm(1, mean  = as.numeric(mu2), sigma = as.matrix(Sigma2),lower = -as.numeric(M*ones(length(mu2),1)), upper = as.numeric(M*ones(length(mu2),1))))

  return(beta_hat)
}

if (sys.nframe() == 0){ #if name == main
  library('pracma')
  library('tmvtnorm')
  library(Matrix)
  source("generate data.R")
  
  #setting constants
  K              = 3     # number of events
  n              = 500    #number of groups
  Egs            = 2
  group_sizes    = sapply(rpois(n, Egs), function(x){max(x,2)})
  niter          = 10000                   # length of Markov chain
  burnin         = 500                    # burn-in period Markov chain
  deg            = 4                      # degree of the I-splines
  p              = 5                      #num covariates
  rho_true       = runif(1, -1/K , 0.3)
  tau_true       = runif(1, -1/(K*max(group_sizes)) - rho_true/max(group_sizes), 0.3)
  dknot         = 0.3
  alpha_coefs   = c(-5, runif(1,10/log(10),20/log(10)), runif(1,0,3))
  alpha_true    = function(t){alpha_coefs[1] +alpha_coefs[2]*log(t) }
  
  #generate data, whole numbers would mean that the endpoints are integers
  data          = generate_data(K, n, group_sizes , tau_true, rho_true, alpha_true, p)
  
  L           = data$L        # left interval endpoints
  R           = data$R       # right interval endpoints
  X           = data$X        # covariate values
  group_sizes = data$group_sizes
  beta_true   = data$beta
  alpha_true  = data$alpha
  Z_vec      = matrix(data$Z, ncol = 1)
  
  mu_0 = zeros(p,1)
  Lambda_0 = zeros(p)
  
  #calculating matrices with which to transform the data later on
  cugs                  = c(0, cumsum(group_sizes))
  cluster_indc_mat      = bdiag(lapply(group_sizes, function(gs){ones(gs*K,1)}))
  subject_indc_mat      = bdiag(lapply(seq(sum(group_sizes)), function(i){ones(K,1)}))

  sum_X_clusters        = t(cluster_indc_mat)%*%X   
  sum_X_subjects        = t(subject_indc_mat)%*%X
  subject_coef  = -rho_true/(K*rho_true + 1)
  clustercoef_matrix = diag(-tau_true/((K*rho_true + 1)*(1+K*(rho_true + group_sizes*tau_true))))
  M = 1000
  
  #use MCMC sampler to sample results
  niter = 1000
  MCMCres = c()
  for(i in seq(niter)){
    print(i)
  MCMCres = rbind(MCMCres, t(sample_beta(X,Z_vec, clustercoef_matrix , sum_X_clusters , 
                                         cluster_indc_mat, subject_coef , sum_X_subjects , subject_indc_mat , mu_0, Lambda_0, M)))
  }

  #plot results
  matplot(cbind(MCMCres, ones(M,1)%*%beta_true), type = 'l', lwd = 2)
}