
balance_data = function(m_max, group_sizes, K, rho_hat, tau_hat, suffstats_tau){
  # simulate the remaining subject sums to artifiically balance the data.
  n = length(group_sizes)
  m_resid             = m_max - group_sizes
  vars_suffstats_aug  = 1+K*rho_hat+K*m_resid*tau_hat - K^2*tau_hat^2*group_sizes*m_resid/(1+K*rho_hat + +K*group_sizes*tau_hat)
  means_suffstats_aug = K*tau_hat*sqrt(group_sizes*m_resid)/(1+K*rho_hat + K*group_sizes*tau_hat)*suffstats_tau
  suffstats_aug       = rnorm(n, means_suffstats_aug, sqrt(vars_suffstats_aug))
  SSB_aug_tau         = sum((suffstats_tau*sqrt(K*group_sizes)+ suffstats_aug*sqrt(K*m_resid))^2/(K*m_max))
  return(SSB_aug_tau)
}


sample_covariances = function(suffstats_tau, suffstats_rho, tau_hat, rho_hat, group_sizes, K,a_tau=0, b_tau = 0, a_rho = 0, b_rho = 0){
  # Gibbs sampler for sampling covariance parameters a posteriori assuming normal data with differing variance depending on tau and rho and parameters that are distributed according to a 
  # shifted inverse gamma distribution.
  # model: let suffstats_rho = c(S_{11}, ..., S_{N1}) and suffstats_tau=c(S_{12},...,S_{M2}) and assume 
  # S_{i1} ~ N(0, 1+K*rho) for all i
  # S_{i2} ~ N(0, 1+K*rho + K*n_{i2}*tau)
  
  # given the current iterates tau_hat, rho_hat, the group_sizes vector c(n_{12},...,n_{M2}), K and the shifted inverse gamma prior scale and shape parameters a_tau,b_tau and a_rho,b_rho for tau and rho respectively
  # the method then samples (tau, rho) using data augmentation.
  
  #determine constants
  n           = length(group_sizes)
  m_max       = max(group_sizes)
  
  #balance the sufficient statistics
  SSB_aug_tau = balance_data(m_max, group_sizes, K, rho_hat ,tau_hat, suffstats_tau)
  
  # sample K*m_max*tau + 1 + K*rho
  stat_covs_1 = rinvgamma(1, a_tau + n/2, b_tau + SSB_aug_tau/2)
  
  # sample K*rho+1
  shape = a_rho + (sum(group_sizes)-n)/2
  rate  = b_rho + sum(suffstats_rho^2)/2
  stat_covs_2 = rinvgamma(1, shape,  rate)
  
  #determine tau, rho from these statistics
  rho_hat = (stat_covs_2 - 1)/K
  tau_hat = (stat_covs_1 - K*rho_hat - 1)/(K*m_max)
  
  return(c(tau_hat, rho_hat))
}

if (sys.nframe() == 0){ #if name == main, example to check sampling method

  library('pracma')
  library('mvtnorm')
  library(Matrix)
  library(invgamma)

  source("generate data.R")
  source("utils.R")
  #setting constants
  K              = 5     # number of events
  n              = 300    #number of groups
  Egs            = 5
  m_max          = 10
  group_sizes    = sapply(rpois(n, Egs), function(x){min(m_max, max(x,2))})
  niter          = 10000                   # length of Markov chain
  burnin         = 500                    # burn-in period Markov chain
  deg            = 4                      # degree of the I-splines
  p              = 5                      #num covariates
  tau_true       = -0.5#runif(1,-0.5, 0.5)
  rho_true       = -1/K + max(0, -m_max*tau_true) + runif(1,0,1)
  dknot         = 0.3
  alpha_coefs   = c(-5, runif(1,10/log(10),20/log(10)), runif(1,0,3))
  alpha_true    = function(t){alpha_coefs[1] +alpha_coefs[2]*log(t) }

  #generate data, whole numbers would mean that the endpoints are integers
  data          = generate_data(K, n, group_sizes , tau_true, rho_true, alpha_true, p)

  L           = data[[1]]        # left interval endpoints
  R           = data[[2]]        # right interval endpoints
  X           = data[[3]]        # covariate values
  group_sizes = data[[4]]
  beta_true   = data[[5]]
  alpha_true  = data[[6]]
  Z_vec       = data[[9]]

  #determining matrices with which one can sample tau,rho
  cugs = c(0, cumsum(group_sizes))
  Helmerttransformmat   = bdiag(lapply(group_sizes, function(gs){kron(HelmertMatrix(gs), HelmertMatrix(K))}))
  start_cluster_indices = cugs[seq(n)]*K+1
  other_subject_indices = setdiff(seq(1,sum(group_sizes)*K,K), start_cluster_indices)
  suffstat_tau_mat      = Helmerttransformmat[start_cluster_indices,]
  suffstat_rho_mat      = Helmerttransformmat[other_subject_indices,]

  # sampling tau
  V_hat         = Z_vec - X%*%beta_true
  suffstats_tau = as.numeric(suffstat_tau_mat%*%V_hat)
  suffstats_rho = as.numeric(suffstat_rho_mat%*%V_hat)

  tau_hat = 0
  rho_hat = 0
  M = 1000
  burnin = 100
  MCMC_results = c()
  for(i in seq(M)){
    res = sample_covariances(suffstats_tau, suffstats_rho, tau_hat, rho_hat, group_sizes, K)
    tau_hat = res[1]
    rho_hat = res[2]
    MCMC_results = rbind(MCMC_results, res)
  }
  MCMC_results = MCMC_results[seq(burnin+1,M),]
  print(summary(MCMC_results))
  matplot(cbind(MCMC_results, ones(M-burnin,1)*tau_true, ones(M-burnin,1)*rho_true), type = 'l', col = c('black','red','blue','green'))
  legend('center',c('tau_hat','rho_hat','tau','rho'),col = c('black','red','blue','green'),
         lty = c(1,1,2,2))


}
