library(Matrix)


load_obj <- function(f)
  #https://stackoverflow.com/questions/5577221/how-can-i-load-an-object-into-a-variable-name-that-i-specify-from-an-r-data-file
{
  env <- new.env()
  nm <- load(f, env)[1]
  env[[nm]]
}


trim_data = function(data, level = 0.01){
   data = data[data<quantile(data,1-level/2)]
   data[data>quantile(data, level/2)]
}

get_spline_matrix_points = function(evpoints, knots, deg){
  #returns the spline matrix at evaluation points "times" with splines of degree "deg" increasing between "knots".
  
  
  num_funcs   = length(knots) + deg-1
  indc_0      = evpoints == 0
  sum_0       = sum(indc_0)
  indc_inf    = evpoints == Inf
  num_inf     = sum(indc_inf)
  indc_normal = !(indc_inf | indc_0)
  res         = zeros(length(evpoints), num_funcs+1)
  
  res[indc_0,]                        = -c(zeros(sum_0,1), ones(sum_0, num_funcs)*Inf)
  res[indc_inf,]                      =  c(zeros(num_inf,1), ones(num_inf,num_funcs)*Inf)

    if(length(knots)>2){
      res[indc_normal,seq(2,num_funcs+1)] = iSpline(evpoints[indc_normal], knots = knots[seq(2, length(knots)-1)], degree = deg,
                                                    intercept = TRUE, Boundary.knots = c(knots[1], knots[length(knots)]))
      
    }else{
      res[indc_normal,seq(2,num_funcs+1)] = iSpline(evpoints[indc_normal], degree = deg,
                                                    intercept = TRUE, Boundary.knots = c(knots[1], knots[length(knots)]))
    }

   res[indc_normal,1]                  = ones(sum(indc_normal),1) # adding intercept term
   
   
  return(res)
}

get_spline_function = function(knots, degree, coefs){
  if(!(length(coefs) == length(knots) + degree )){
    warning(paste('WARNING: length of coefs should be',toString(length(knots) + degree)))
    return(NULL)
  }else{
    return(Vectorize(function(t){
      if(t<=0){return(-Inf)}else{
        if(t==Inf){return(Inf)}
        if(t<min(knots)){
          return(coefs[1])
        }else{
          if(t>max(knots)){
            return(sum(coefs))
          }else{
            if(length(knots) == 2){
              return(coefs[1] + sum(iSpline(t, degree = deg,
                                            intercept = TRUE, Boundary.knots = c(knots[1], knots[length(knots)]))*t(coefs[seq(2,length(coefs))])))
              
            }else{
            return(coefs[1] + sum(iSpline(t, knots = knots[seq(2, length(knots)-1)], degree = deg,
                                          intercept = TRUE, Boundary.knots = c(knots[1], knots[length(knots)]))*t(coefs[seq(2,length(coefs))])))
         } }}}}))}}

HelmertMatrix = function(n){
  #obtain Helmert matrix of dimension nxn
  if(n==1){return(1)}else{
    result = ones(1,n)/sqrt(n)
    for(i in seq(2,n)){
      result = rbind(result, c(ones(1,i-1)/sqrt(i*(i-1)), -(i-1)/sqrt(i*(i-1)), zeros(1,n-i)))
    }
    return(result)
  }}

Diagmat = function(vec){
  if(length(vec)==1){return(vec)}else{
    return(diag(as.numeric(vec)))
  }
}

rtnorm_custom = function(num, mean = 0, sd = 1, a = -Inf, b = Inf){
  if(length(sd)==1){
    sd = sd*ones(num,1)
  }
  lb_prob = pnorm((a-mean)/sd)
  ub_prob = pnorm((b-mean)/sd)
  res = as.numeric(NA*ones(num,1))
  
  Indc_extreme_rstr     = (lb_prob == 1)&(b<Inf) | (ub_prob == 0)&(a>-Inf) 
  Indc_too_large_unrstr = (lb_prob == 1)&(b==Inf)
  Indc_too_small_unrstr = (ub_prob == 0)&(a==-Inf)
  Indc_nontrunc         = (b == Inf) & (a == -Inf)
  Indc_trunc            = !(Indc_nontrunc| Indc_extreme_rstr | Indc_too_large_unrstr | Indc_too_small_unrstr)
  
  if(sum(Indc_extreme_rstr) >0){res[Indc_extreme_rstr] = runif(sum(Indc_extreme_rstr), a[Indc_extreme_rstr],b[Indc_extreme_rstr])}
  if(sum(Indc_too_large_unrstr) >0){res[Indc_too_large_unrstr] = a[Indc_too_large_unrstr]}
  if(sum(Indc_too_small_unrstr) >0){res[Indc_too_small_unrstr] = b[Indc_too_small_unrstr]}
  if(sum(Indc_nontrunc) >0){res[Indc_nontrunc] = rnorm(sum(Indc_nontrunc), mean= mean[Indc_nontrunc], sd = sd[Indc_nontrunc])}
  if(sum(Indc_trunc) >0){res[Indc_trunc] = qnorm(runif(sum(Indc_trunc), lb_prob[Indc_trunc], ub_prob[Indc_trunc]))*sd[Indc_trunc] + mean[Indc_trunc]}
  
  return(res)
}


get_row = function(mat,rownr){
  if(is.null(dim(mat))){
    return(mat[rownr])
  }else{
    return(mat[rownr,])
  }
}

set_row = function(mat, rownr, row){
  if(is.null(dim(mat))){
    mat[rownr] = row
  }else{
    mat[rownr,] = row
  }
  return(mat)
}


mult_sqrtinvmat = function(tau, rho, m, K, v){
  return( v + 1/(m*K)*(1/sqrt(1+K*rho+m*K*tau) - 1/sqrt(1+K*rho))*mean(v) + 1/K*(1/sqrt(1+K*rho)-1)*bdiag(lapply(seq(m), function(i){sum(v[seq(K*(i-1)+1, K*i)])*ones(K)})))
}

